// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net;

using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Client.Transaction;
using Smdn.Net.Imap4.Protocol.Client;
using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Imap4.Client {
  public class ImapClient : IDisposable {
    public ImapClientProfile Profile {
      get { return profile; }
    }

    public int Timeout {
      get { return (session == null) ? profile.Timeout : session.TransactionTimeout; }
      set
      {
        profile.Timeout = value;

        if (session != null)
          session.TransactionTimeout = value;
      }
    }

    public int SendTimeout {
      get { return (session == null) ? profile.SendTimeout : session.SendTimeout; }
      set
      {
        profile.SendTimeout = value;

        if (session != null)
          session.SendTimeout = value;
      }
    }

    public int ReceiveTimeout {
      get { return (session == null) ? profile.ReceiveTimeout : session.ReceiveTimeout; }
      set
      {
        profile.ReceiveTimeout = value;

        if (session != null)
          session.ReceiveTimeout = value;
      }
    }

    public bool IsConnected {
      get { return (session != null && session.State != ImapSessionState.NotConnected); }
    }

    internal ImapSession Session {
      get { CheckConnected(); return session; }
    }

    public ImapCapabilityList ServerCapabilities {
      get { return Session.ServerCapabilities; }
    }

    public IDictionary<string, string> ServerID {
      get { return Session.ServerID; }
    }

    public ImapNamespace ServerNamespace {
      get { return Session.Namespaces; }
    }

    /*
     * construction
     */
    public ImapClient()
    {
      profile = new ImapClientProfile();
    }

    public ImapClient(Uri authority)
    {
      profile = new ImapClientProfile(authority);
    }

    public ImapClient(string host, string userName)
      : this()
    {
      profile.Host = host;
      profile.UserName = userName;
    }

    public ImapClient(string host, int port, string userName)
      : this()
    {
      profile.Host = host;
      profile.Port = port;
      profile.UserName = userName;
    }

    public ImapClient(string host, int port, string userName, string authType)
      : this()
    {
      profile.Host = host;
      profile.Port = port;
      profile.UserName = userName;
    }

    public void Connect(string password)
    {
      Connect(password, null);
    }

    public void Connect(string password, UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      Connect(new NetworkCredential(profile.UserName, password), createSslStreamCallback);
    }

    public void Connect(ICredentialsByHost credentials)
    {
      Connect(credentials, null);
    }

    public void Connect(ICredentialsByHost credentials, UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      try {
        profile.SetCredentials(credentials);

        ThrowIfError(ImapSessionCreator.CreateSession(profile,
                                                      createSslStreamCallback ?? ImapConnection.CreateSslStream,
                                                      out session));

        // update server info
        if (session.ServerCapabilities.Has(ImapCapability.Namespace))
          session.Namespace();

        if (session.ServerCapabilities.Has(ImapCapability.ID))
          session.ID(null); // TODO: client ID
      }
      finally {
        profile.SetCredentials(null);
      }
    }

    void IDisposable.Dispose()
    {
      Disconnect(false);
    }

    public void Disconnect()
    {
      Disconnect(false);
    }

    public void Logout()
    {
      Disconnect(true);
    }

    public void Disconnect(bool logout)
    {
      if (session != null)
        session.Disconnect(logout);

      session = null;
    }

    private ImapSession session;
    private ImapClientProfile profile;

    /*
     * operations
     */
    private ImapOpenedMailboxInfo openedMailbox;

    public ImapOpenedMailboxInfo OpenedMailbox {
      get { CheckConnected(); return openedMailbox; }
    }

    public ImapMailboxInfo GetInbox()
    {
      return GetMailbox(ImapMailbox.Inbox);
    }

    public ImapMailboxInfo GetMailbox(string mailboxName)
    {
      ImapMailbox[] mailboxes;

      ThrowIfError(Session.List(GetValidatedMailboxName(mailboxName), out mailboxes));

      if (mailboxes.Length == 0)
        throw new ImapException(string.Format("mailbox not found: '{0}'", mailboxName));

      return new ImapMailboxInfo(this, mailboxes[0]);
    }

    public IEnumerable<ImapMailboxInfo> GetMailboxes()
    {
      return GetMailboxes(ImapMailboxListOptions.Default);
    }

    public IEnumerable<ImapMailboxInfo> GetMailboxes(ImapMailboxListOptions options)
    {
      return GetMailboxes(null, null, false, options);
    }

    internal IEnumerable<ImapMailboxInfo> GetMailboxes(ImapMailbox mailbox,
                                                       ImapMailboxListOptions options)
    {
      return GetMailboxes(mailbox.Name, mailbox.HierarchyDelimiter, mailbox.IsNonHierarchical, options);
    }

    internal IEnumerable<ImapMailboxInfo> GetMailboxes(string mailboxName,
                                                       string hierarchyDelimiter,
                                                       bool isNonHierarchical,
                                                       ImapMailboxListOptions options)
    {
      string mailboxNamePattern;

      if (mailboxName == null) {
        mailboxNamePattern = string.Empty;
      }
      else {
        if (string.IsNullOrEmpty(hierarchyDelimiter))
          yield break;
        else if (isNonHierarchical)
          yield break;

        mailboxNamePattern = mailboxName + hierarchyDelimiter;
      }

      mailboxNamePattern += ((int)(options & ImapMailboxListOptions.TopLevelOnly) != 0)
        ? "%"
        : "*";

      var subscribedOnly = (int)(options & ImapMailboxListOptions.SubscribedOnly) != 0;
      var listRemote = (int)(options & ImapMailboxListOptions.Remote) != 0;
      var requestStatus = (int)(options & ImapMailboxListOptions.RequestStatus) != 0;

      ImapMailbox[] mailboxes;

      if (Session.ServerCapabilities.Has(ImapCapability.ListExtended)) {
        var selectionOptions = ImapListSelectionOptions.Empty;
        var returnOptions = ImapListReturnOptions.Children;

        if (subscribedOnly)
          selectionOptions += ImapListSelectionOptions.Subscribed;

        if (listRemote)
          selectionOptions += ImapListSelectionOptions.Remote;

        if (requestStatus) {
          returnOptions += ImapListReturnOptions.StatusDataItems(ImapStatusDataItem.StandardAll);
          requestStatus = false; // no need to request STATUS
        }

        ThrowIfError(Session.ListExtended(mailboxNamePattern,
                                          selectionOptions,
                                          returnOptions,
                                          out mailboxes));
      }
      else if (Session.ServerCapabilities.Has(ImapCapability.MailboxReferrals) &&
               listRemote) {
        if (subscribedOnly)
          ThrowIfError(Session.RLsub(mailboxNamePattern, out mailboxes));
        else
          ThrowIfError(Session.RList(mailboxNamePattern, out mailboxes));
      }
      else {
        if (subscribedOnly)
          ThrowIfError(Session.Lsub(mailboxNamePattern, out mailboxes));
        else
          ThrowIfError(Session.List(mailboxNamePattern, out mailboxes));
      }

      foreach (var m in mailboxes) {
        if (requestStatus && !m.IsUnselectable)
          ThrowIfError(Session.Status(m, ImapStatusDataItem.StandardAll));

        yield return new ImapMailboxInfo(this, m);
      }
    }

    public ImapMailboxInfo CreateMailbox(string mailboxName)
    {
      return CreateMailbox(mailboxName, false);
    }

    public ImapMailboxInfo CreateMailbox(string mailboxName, bool subscribe)
    {
      ImapMailbox createdMailbox;

      ThrowIfError(Session.Create(GetValidatedMailboxName(mailboxName), out createdMailbox));

      if (subscribe)
        ThrowIfError(Session.Subscribe(createdMailbox));

      return new ImapMailboxInfo(this, createdMailbox);
    }

    public ImapOpenedMailboxInfo OpenInbox()
    {
      return OpenMailbox(ImapMailbox.Inbox, false);
    }

    public ImapOpenedMailboxInfo OpenInbox(bool readOnly)
    {
      return OpenMailbox(ImapMailbox.Inbox, readOnly);
    }

    public ImapOpenedMailboxInfo OpenMailbox(ImapMailboxInfo mailbox)
    {
      return OpenMailbox(mailbox.FullName, false);
    }

    public ImapOpenedMailboxInfo OpenMailbox(ImapMailboxInfo mailbox, bool readOnly)
    {
      if (!mailbox.Exists)
        throw new InvalidOperationException("mailbox is not existent");
      else if (mailbox.IsUnselectable)
        throw new InvalidOperationException("mailbox is not selectable");

      return OpenMailbox(mailbox.FullName, readOnly);
    }

    public ImapOpenedMailboxInfo OpenMailbox(string mailboxName)
    {
      return OpenMailbox(mailboxName, false);
    }

    public ImapOpenedMailboxInfo OpenMailbox(string mailboxName, bool readOnly)
    {
      if (openedMailbox != null)
        openedMailbox.SetClosed();

      lock (Session) {
        ThrowIfError(readOnly
                     ? Session.Examine(GetValidatedMailboxName(mailboxName))
                     : Session.Select(GetValidatedMailboxName(mailboxName)));

        openedMailbox = new ImapOpenedMailboxInfo(this, Session.SelectedMailbox);

        return openedMailbox;
      }
    }

    public void CloseMailbox()
    {
      CloseMailbox(openedMailbox);
    }

    internal void CloseMailbox(ImapOpenedMailboxInfo mailbox)
    {
      if (mailbox != openedMailbox)
        throw new InvalidOperationException("mailbox is already closed");

      lock (Session) {
        ThrowIfError(Session.Close());

        this.openedMailbox = null;
      }
    }

    public void NoOp()
    {
      ThrowIfError(Session.NoOp());
    }

    internal static void ThrowIfError(ImapTransactionResult result)
    {
      if (result.Failed)
        throw new ImapException(result.Description);
    }

    internal static string GetValidatedMailboxName(string mailboxName)
    {
      if (mailboxName == null)
        throw new ArgumentNullException("mailboxName");
      else if (mailboxName.Length == 0)
        throw new ArgumentException("empty mailbox name is not allowed", "mailboxName");
      else if(mailboxName.Contains("%") || mailboxName.Contains("*"))
        throw new ArgumentException("mailbox name with wildcards is not allowed", "mailboxName");

      return mailboxName;
    }

    /*
     * methods for checking and validating internal status
     */
    private void CheckConnected()
    {
      if (!IsConnected)
        throw new InvalidOperationException("not connected or disconnected");
    }
  }
}
