// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Text;

using Smdn.Collections;

namespace Smdn.Net.Imap4.Client {
  public class ImapOpenedMailboxInfo : ImapMailboxInfo, IDisposable {
    internal ImapOpenedMailboxInfo(ImapClient client, ImapMailbox mailbox)
      : base(client, mailbox)
    {
    }

    void IDisposable.Dispose()
    {
      this.Close();
    }

    public void Close()
    {
      Client.CloseMailbox(this);
    }

    internal void SetClosed()
    {
      closed = true;
    }

    internal void CheckClosed()
    {
      if (closed)
        throw new InvalidOperationException("mailbox has been closed");
    }

    private bool closed = false;

    /*
     * operations
     */
    public override void Refresh()
    {
      ImapClient.ThrowIfError(Client.Session.NoOp());
    }

    private const long fetchCount = 100L;

    public IEnumerable<ImapMessageInfo> GetMessages()
    {
      return GetMessages(ImapMessageFetchOptions.Default);
    }

    public IEnumerable<ImapMessageInfo> GetMessages(ImapMessageFetchOptions options)
    {
      CheckClosed();

      Refresh();

      if (Mailbox.ExistsMessage <= 0L)
        yield break;

      var fetchStaticAttr = (int)(options & ImapMessageFetchOptions.FetchStaticAttribute) != 0;
      var fetchDynamicAttr = (int)(options & ImapMessageFetchOptions.FetchDynamicAttribute) != 0;
      var fetchDataItem = ImapFetchDataItem.Uid;

      if (fetchStaticAttr)
        fetchDataItem += ImapMessageInfo.FetchStaticDataItem;
      if (fetchDynamicAttr)
        fetchDataItem += ImapMessageInfo.GetFetchDynamicDataItem(Client.ServerCapabilities);

      for (var sequenceFrom = 1L;;) {
        if (Mailbox.ExistsMessage < sequenceFrom)
          yield break;

        var sequenceTo = Math.Min(sequenceFrom + fetchCount - 1, Mailbox.ExistsMessage);
        var sequenceSet = (sequenceFrom == sequenceTo)
          ? ImapSequenceSet.CreateSet(sequenceFrom)
          : ImapSequenceSet.CreateRangeSet(sequenceFrom, sequenceTo);

        ImapMessageAttribute[] messages;

        ImapClient.ThrowIfError(Client.Session.Fetch(sequenceSet,
                                                     fetchDataItem,
                                                     out messages));

        foreach (var message in messages) {
          var info = new ImapMessageInfo(Client, this, message.Uid, Mailbox.UidValidity, message.Sequence);

          if (fetchStaticAttr)
            info.StaticAttribute = message.GetStaticAttributeImpl();
          if (fetchDynamicAttr)
            info.DynamicAttribute = message.GetDynamicAttributeImpl();

          yield return info;
        }

        sequenceFrom += fetchCount;
      }
    }

    public ImapMessageInfo GetMessageByUid(long uid)
    {
      return GetMessages(ImapSearchCriteria.Uid(ImapSequenceSet.CreateUidSet(uid))).FirstOrDefault();
    }

    public IEnumerable<ImapMessageInfo> GetMessages(ImapSearchCriteria searchCriteria)
    {
      return GetMessages(searchCriteria, null, ImapMessageFetchOptions.Default);
    }

    public IEnumerable<ImapMessageInfo> GetMessages(ImapSearchCriteria searchCriteria, ImapMessageFetchOptions options)
    {
      return GetMessages(searchCriteria, null, options);
    }

    public IEnumerable<ImapMessageInfo> GetMessages(ImapSearchCriteria searchCriteria, Encoding encoding)
    {
      return GetMessages(searchCriteria, encoding, ImapMessageFetchOptions.Default);
    }

    public IEnumerable<ImapMessageInfo> GetMessages(ImapSearchCriteria searchCriteria, Encoding encoding, ImapMessageFetchOptions options)
    {
      if (searchCriteria == null)
        throw new ArgumentNullException("searchCriteria");

      ImapMatchedSequenceSet matchedSequenceNumbers;

      ImapClient.ThrowIfError(Client.Session.Search(searchCriteria, encoding, out matchedSequenceNumbers));

      foreach (var sequenceSet in GetBatchedSequenceSet(matchedSequenceNumbers, fetchCount)) {
        ImapMessageStaticAttribute[] messages;

        ImapClient.ThrowIfError(Client.Session.Fetch(sequenceSet,
                                                     ImapFetchDataItem.Uid,
                                                     out messages));

        foreach (var message in messages) {
          yield return new ImapMessageInfo(Client, this, message.Uid, Mailbox.UidValidity, message.Sequence);
        }
      }
    }

    private IEnumerable<ImapSequenceSet> GetBatchedSequenceSet(ImapSequenceSet baseSequenceOrUidSet, long count)
    {
      var batched = new List<long>();

      foreach (var num in baseSequenceOrUidSet) {
        batched.Add(num);

        if (count <= batched.Count) {
          yield return ImapSequenceSet.CreateSet(baseSequenceOrUidSet.IsUidSet, batched.ToArray());

          batched.Clear();
        }
      }

      if (0 < batched.Count)
        yield return ImapSequenceSet.CreateSet(baseSequenceOrUidSet.IsUidSet, batched.ToArray());
    }
  }
}
